/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          vhi version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var iter = 2;   // Number of iterations for each variable value

var line_size = [];   // Line size in percents for each item ( 0..item_cnt-1 ) in task

var back_clr = '#fff';  // background canvas color during task

var pen_clr = '#000';  //  font color for a line that is fixed 

var pen_clr_m = '#000';  // font color for moving line (when is not-dragging) 

var drag_clr = '#f00'; //  font color for a scrolling line when dragged with the mouse

var cor_clr = '#006400';  // Font color to display the correct solution

var fixed_line = '1';     //  The fixed line is: Vertical == '1' or Horizontal == '2'

var pen_width = 3;     // Pen width for all lines
  
var is_back = true;    // if (is_back==true) => user can go back to previous task/item.  

var prg_type = true;   //  Progress Bar is visible (true) or invisible (false)

var demo = false;       // if (demo == true) => Show task solutions 

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global constants and variables:

var item_cnt;       // number of items in the task. range [0..item_cnt-1]

var item;   // current item in task

var is_task = false;    // task in progress or not for pageResize() event. It is true for all events between end of startTask() and theEnd() events

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_storage = false; // local storage is enabled on user browser or not

var is_move = false, start_rad;  // dragging var. see: onMouseDown(), onMouseUp(), mouseMove(),..

var r_max_prior;  // last used max canvas radius in stimGen(). used if canvas width or height changed during task 			 

var resized = 0;  // number of resized events during task 
  
//  global objects and arrays declarations..
 
var stim = {};   // stimulus object. see: stimGen()  

var dv = [];  // dependent variable for each item in range (error-pixels and percents). dimensions: [0 .. item_cnt-1] * [0 == absolute error(px), 1 == relative error in approx range -50..50]. 

var mis = [];  // missing values for each item in range: [0..item_cnt-1]
	

/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


// application init [check storage, set default values, show start panel..] 
function init() 
{ 
  setDefault();	      
  
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }	
	
	document.onkeydown = onKeyDown;
	setIns();
	showStart(); 
}

 
// set default values in <textarea> object 
function setDefault()
{ 
	document.getElementById('line_size').value= '90\n60\n30';
	 //document.getElementById('line_size').value= '90\r\n60\r\n30'; 
}


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 (probably redundant)
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if ( is_task )  { 
    resized++;
		setCanvas();
		stimGen();
		drawStim();		
	}
}	


// show start panel 
function showStart()  
{     
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('vhi-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('vhi-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // xxx
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
	refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		
  document.getElementById('ins_back').style.display= is_back ? 'none' : 'block';
	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty = true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'iter':
			if (e.validity.valid)  {
				msg.textContent= 'Total number of items = (number of iterations) * (number of entered valid values ​​for the line-length)'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 
				
    case 'back_clr': 		// background canvas color during task
		case 'pen_clr':  		//  font color (base font) during the task 
		case 'pen_clr_m': 	// font color for moving line 
		case 'drag_clr':		//  font color for a scrolling line when dragged with the mouse   
    case 'cor_clr':	  // font color to display the correct solution   
			msg.textContent= ( validColor(e.value) ) ? '' : 'The selected color is invalid.'; 
			break; 		
			
		case 'fixed_line':  // The fixed line is: Vertical == '1' or Horizontal == '2'
      msg.textContent= (e.value == '1') ? 'The vertical line is fixed. The horizontal line is movable.' : 'The horizontal line is fixed. The vertical line is movable.';
			break;	
						
    case 'pen_width':
			if (e.validity.valid)  {
				msg.textContent= 'The larger the width of the pen, the larger and clearer all stimulus lines are.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 
			
		case 'is_back':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break;  	
		
    case 'prg_type':                                                                                                                                    
      msg.textContent= e.checked ? 'Yes. The number of the current item will be displayed, as well as the total number of items.' : 'No.';      
      break;    

		case 'demo': 
			msg.textContent= e.checked ? 'On. The solution of each task will be displayed. The error size will be displayed.' : 'Off';  
			break; 
		
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity in most of options (except line size textarea elements)
function optionsValid()  
{ 
  let res = document.getElementById('iter').validity.valid && document.getElementById('pen_width').validity.valid  && validColor(document.getElementById('back_clr').value) && validColor(document.getElementById('pen_clr').value) && validColor(document.getElementById('pen_clr_m').value) && validColor(document.getElementById('drag_clr').value) && validColor(document.getElementById('cor_clr').value);
  return res;              
}


// start task (research) 
function startTask(e)   
{      			
  initLayout(); 	
	
  setPar();
 
	stimGen();
	
	is_move= false;
	is_task= true;  // after call stimGen()
	
	drawStim();    // draw first item (stimulus)
}


// init DOM/html elements (positions,visibility,..)
function initLayout()
{	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block'; 	  	
	document.getElementById('progress_holder').style.display = document.getElementById('prg_type').checked ? 'block' : 'none';
  document.getElementById('nav_items').textContent= '';	
	
  setCanvas();
}  


// set html canvas element layout and absolute positions
function setCanvas()
{
	// canvas must be after set footer content => progress_holder visible or not
	let c= document.getElementById('can'), ctx= c.getContext('2d');
	c.style.display= 'block';   
	c.width= innerWidth;  
	c.style.width= c.width + 'px';	     
	c.height= innerHeight - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight  
	c.style.height= c.height + 'px';  
	
	ctx.font = '1em Arial';   
	if (ctx.font.fontFamily)  {
			ctx.font.fontFamily = 'Arial, Verdana, Helvetica, sans-serif';  // https://www.w3schools.com/cssref/css_websafe_fonts.asp
	}
	ctx.textAlign = 'middle';
	ctx.textBaseline= 'bottom';	
  ctx.lineWidth= pen_width; 	
}
	
	
/* set internal application research parameters (global vars)
   prepare stimulus properties (line size) for stimulus generator  */
function setPar()  
{  
	item = 0; 
	resized= 0;
	
	stim = {};	
	dv = [];
	mis = []; 		
	
	// update global window.line_size var. it is different from getElementById('line_size') html object
	line_size = [];
	
	let size = [], v;	
	let lines = document.getElementById('line_size').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{  // Validate Line size. Enter only one integer value in the range of 10-100 per line 
		lines[i]= trim(lines[i]);
		if ( isInt( lines[i]) )  {  
		  v = parseInt(lines[i], 10);
			if ( v >= 10  &&  v <= 100 )  {
				size.push(v);
			}
		}
	}
	if (size.length == 0)  {
		size.push(100);  // Default value = 100
	}	
	document.getElementById('line_size').value= size.join('\n');	

	// prepare stimulus properties (line size) for stimulus generator	
	size = [];  // do job again.  reset fill size array again from filtered (validated) line size list
	lines = document.getElementById('line_size').value.split('\n');
	for (let i = 0; i < lines.length; i++)  {
		v = parseInt(lines[i], 10);
		for (let j= 1; j<= iter; j++) {  // multiply by number of iterations
			size.push(v);
		}
	}
	if  (size.length > 1) {
		size= shuffle(size);
	}

	item_cnt= size.length;  // number of items in task
	for (let i=0; i < item_cnt; i++)  {
		line_size[i] = size[i];  // user eneterd values in percents. range: [10-100]
		mis[i]= true;	
	}			
}


// https://varun.ca/polar-coords/
// https://tutorial.math.lamar.edu/classes/calcii/polarcoordinates.aspx
// generate all stimuli for each item in range 0..item_cnt-1	
function stimGen()	
{			 		
	const margin = 5;  // offset from edge. canvas position is set in end of initLayout() => after set footer content
	let c= document.getElementById('can'), w= c.width, h= c.height;				
	let r_max = Math.min(w, h) - 2 * margin;
	let fixed_v = fixed_line == '2' ? false: true;  // window.fixed_line: Vertical == '1' or Horizontal == '2'
	for (let i = 0; i < item_cnt; i++)    // i => item  
	{			
		let rad = r_max * line_size[i] / 100, m_rad;     // rad => radius for this item depending of line_size for this item				  
		if (is_task)  {  // m_rad is polar radius for moved line
      m_rad = stim[i].moved.radius * (r_max / r_max_prior);  // load existing (prior) moved line radius. rescale polar radius if canvas width or height changed during task			
		}
		else  {  // first time. call from startTask()
			if ( fixed_v ) {
				m_rad= demo ? rad/2 : rad/4;    // if !demo moved line have cca 50% of fixed line size. else 100% 
			}
			else  {
			  m_rad= demo ? rad : rad/2; 	    // if !demo moved line have cca 50% of fixed line size. else 100% 
			}
		}		
		let xs =  w / 2;    	 // x-start from half of canvas.width. horizontal-center (text-align:center;)
		let ys = (h - rad)/2;	 // y-start from vertical-center depending of line_size[i] => vertical-align:middle;
		let x0 =  Math.round(xs + rad * Math.cos(Math.PI/2));  // T(x0, y0) is line intersection of fixed and moved line.  
		let y0 =  Math.round(ys + rad * Math.sin(Math.PI/2));  // T(x0, y0) is line intersection of fixed and moved line. 
		stim[i] = {};
		if ( fixed_v )  // window.fixed_line is Vertical ('1')
		{  
		  // fixed line is vertical 
			stim[i].fixed =  { 
				'begin' : { x : xs,  y : ys },
				'end'   : { x : x0,  y : y0 }			
			};
			// correct line is horizontal
			stim[i].correct = {
				'begin' : { x : Math.round(x0 - rad/2 * Math.cos(0)), y : Math.round(y0 - rad/2 * Math.sin(0)) },
				'end'   : { x : Math.round(x0 + rad/2 * Math.cos(0)), y : Math.round(y0 + rad/2 * Math.sin(0)) },
			};	
			// moved line is horizonal
			stim[i].moved = { 
				'begin' : { x : Math.round(x0 - m_rad * Math.cos(0)), y : Math.round(y0 - m_rad * Math.sin(0)) },
				'end'   : { x : Math.round(x0 + m_rad * Math.cos(0)), y : Math.round(y0 + m_rad * Math.sin(0)) }, 
				'radius': m_rad,   // default: 50% or 100% currect radius for mouseMove() see start_rad 
				'angle' : 0     
			};
	  }
		else   //  window.fixed_line is Horizontal  ('2')
		{  
		  // horizontal line is fixed
			stim[i].fixed =  { 
				'begin' : { x : Math.round(x0 - rad/2 * Math.cos(0)), y : Math.round(y0 - rad/2 * Math.sin(0)) },
				'end'   : { x : Math.round(x0 + rad/2 * Math.cos(0)), y : Math.round(y0 + rad/2 * Math.sin(0)) },
			};
			// correct line is vertical
			stim[i].correct = {
				'begin' : { x : xs, y : ys },
				'end'   : { x : x0, y : y0 },
			};		
			// moved line is vertical
			stim[i].moved = { 
				'begin' : { x : x0, y : y0 },
				'end'   : { x : Math.round(x0 - m_rad * Math.cos(Math.PI/2)), y : Math.round(y0 - m_rad * Math.sin(Math.PI/2)) }, 
				'radius': m_rad,   // default: 50% or 100%. currect radius for mouseMove() see start_rad 
				'angle' : Math.PI/2  
			};
		}	
		// common part
	  stim[i].fixed.length= getDistance( stim[i].fixed.begin.x, stim[i].fixed.begin.y, stim[i].fixed.end.x, stim[i].fixed.end.y);
		// add center of moved line. for polar coord
	  stim[i].moved.x0= divLine( stim[i].moved.begin.x, stim[i].moved.end.x, -1);  // half of moved line
		stim[i].moved.y0= divLine( stim[i].moved.begin.y, stim[i].moved.end.y, -1);  // half of moved line
		
		dv[i]= [null,null];	 // reset always ( if is_task == true || is_task == false)

	}  // for (let i = 0; i < item_cnt; i++)
		
  r_max_prior= r_max;   // save last used max radius for possible rescale	
}

 
// isNaN("") = false: the empty string is converted to 0 which is not NaN                   
function isInt(n) {       
   return (!isNaN(n)) && (parseFloat(n) === parseInt(n, 10)) &&  (n % 1 === 0); 
} 


function trim(str) {
   return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}


// unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.
function shuffle(arr) 
{
  let cur_index = arr.length, tmp, rnd_index;
  // While there remain elements to shuffle...
  while (0 !== cur_index) {    
    rnd_index = Math.floor(Math.random() * cur_index);     // Pick a remaining element...
    cur_index = cur_index - 1;   
    tmp = arr[cur_index];          // And swap it with the current element.
    arr[cur_index] = arr[rnd_index];
    arr[rnd_index] = tmp;
  }
  return arr;
}


// https://www.w3.org/2002/09/tests/keys.html	 
// https://stackoverflow.com/questions/17164239/add-delete-and-arrow-key-into-regular-expression 
// https://stackoverflow.com/questions/13196945/keycode-values-for-numeric-keypad/13196983
// textarea validator => onkeydown() => see html tag
function updateVar(event)
{
	event = event || window.event;
	let key = event.keyCode || event.which;
	if( key == 13 || key == 37 || key == 38 || key == 39 || key == 40 || key == 8 || key == 46 ) { // ENTER, Left / Up / Right / Down Arrow, Backspace, Delete keys
		return;  // Don't validate the input if ENTER, arrows, delete and backspace keys were pressed 
	}
	let is_digit = (key >= 48 && key <= 57) || (key >= 96 && key <= 105);
	if( !is_digit )  {
		event.returnValue = false;
		if ( event.preventDefault ) {
			event.preventDefault();
		}  
	}		
}


// return distance from point P1(x1,y1) to point P2(x2,y2)
function getDistance(x1, y1, x2, y2)  {
  return Math.sqrt( Math.pow(x1-x2, 2) + Math.pow(y1-y2, 2) ); // total line distance (px) 
}	


/* https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
  If the line passes through two points P1 = (x1, y1) and P2 = (x2, y2) then the distance of (x0, y0) from the line is d
  d = distance = (P1, P2, (x0,y0)) 
	d = | (x2 - x1) * (y1 - y0) - (x1 - x0) * (y2 - y1) | / ( SQRT ( (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) )  */
function getDistanceFromLine(x1, y1, x2, y2, x0, y0)
{
	let d = Math.abs( (x2-x1) * (y1-y0) - (x1-x0) * (y2 - y1) ) / Math.sqrt( Math.pow(x2-x1, 2) + Math.pow(y2-y1, 2) );  
	return d;
}	


// html <canvas id='can' onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>                         
function mouseDown(e)  
{	 
	let xm= e.offsetX, ym= e.offsetY, xc, yc;   // mouse pointers. Note: e.offsetX == e.clientX;  e.offsetY == e.clientY 
	let d= getDistanceFromLine( stim[item].moved.begin.x, stim[item].moved.begin.y, stim[item].moved.end.x, stim[item].moved.end.y, xm, ym );   // get mouse distance from moved line. e.offsetX and e.offsetY => mouse coordinates x, y	
	let ok= d <= 10 ? true: false;  // 1. condition
	if ( ok )  {  // and 2. condition 
		let fixed_v = fixed_line == '2' ? false: true;  // window.fixed_line: Vertical == '1' or Horizontal == '2'
		if ( fixed_v )  {  //   The fixed line is: Vertical == '1'
			xc= divLine(stim[item].moved.begin.x, stim[item].moved.end.x, -1);			
			yc= divLine(stim[item].moved.begin.y, stim[item].moved.end.y, -1);						
		}
		else  {   // The fixed line is: Horizontal == '2'
			xc = stim[item].moved.begin.x;  // x-intersection of vertical (moved) and horizontal (fixed) line
			yc = stim[item].moved.begin.y;  // y-intersection of vertical (moved) and horizontal (fixed) line
		}		
		let d = getDistance(xm, ym, xc, yc);  // mouse distance from center of moving line 	
		let length= getDistance(stim[item].moved.begin.x, stim[item].moved.begin.y, stim[item].moved.end.x, stim[item].moved.end.y);		
		if ( fixed_v )  { // window.fixed_line is Vertical == '1' 
			ok = d <= length/2 ? true : false;	// anywhere "near within" line	*/
		}
		else  {  // window.fixed_line is Horizontal == '2'
			ok = d <= length ? true : false;	// anywhere "near within" line	*/
		}
	}
	if ( ok )  {				
		start_rad = getDistance( xm, ym, xc, yc );  // mouse pointer distance from center of horizontal moved line or begin of moved vertical line	
		is_move= true;
		document.getElementById('can').style.cursor= 'move';  //  'ew-resize'  or  'ns-resize'  
		drawStim();		// repaint 
  }
	else  {
		stopDrag();
	}
}


// html <canvas id='can' onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>                        
function mouseUp()  
{
	stopDrag();
  drawStim();
}


// stop dragging moving line
function stopDrag()  
{
	is_move= false;
  document.getElementById('can').style.cursor= 'auto';     //c.style.cursor= 'initial';  // because of IE11
	drawStim();	
}


/* event when user move scrolling line with mouse
   canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
   this event is fired by html canvas element => document.getElementById('can') 
	 <canvas id='can' onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>  */
function mouseMove(e)
{   
	if (is_move)   
	{ 	
		let xm= e.offsetX, ym= e.offsetY, xc, yc;
    let fixed_v = fixed_line == '2' ? false: true;  // window.fixed_line: Vertical == '1' or Horizontal == '2'
		if ( fixed_v )  { 
			xc= divLine(stim[item].moved.begin.x, stim[item].moved.end.x, -1);			
			yc= divLine(stim[item].moved.begin.y, stim[item].moved.end.y, -1);		
		}
		else  {
			xc = stim[item].moved.begin.x;  // x-intersection of vertical (moved) and horizontal (fixed) line
			yc = stim[item].moved.begin.y;  // y-intersection of vertical (moved) and horizontal (fixed) line
		}
		let cur_rad = getDistance ( xm, ym, xc, yc ); 	  // mouse pointer current distance form center of canvas (screen)					
		// calc how far mouse has moved from canvas center since last mousemove event
		let r = stim[item].moved.radius + cur_rad - start_rad;    
		let angle= stim[item].moved.angle;
		let x1, y1, x2, y2;  // test new coord x1, y1, x2, y2
		if ( fixed_v )  { 
			x1 = Math.round (xc - r * Math.cos(angle));
			y1 = Math.round (yc - r * Math.sin(angle));
			x2=  Math.round (xc + r * Math.cos(angle));
			y2=  Math.round (yc + r * Math.sin(angle));
		}
		else  {
		  x1 = xc;  // xc = stim[item].moved.begin.x;   // x-intersection of vertical (moved) and horizontal (fixed) line
			y1 = yc;  // yc = stim[item].moved.begin.y;  // y-intersection of vertical (moved) and horizontal (fixed) line
			x2=  Math.round (xc - r * Math.cos(angle));
			y2=  Math.round (yc - r * Math.sin(angle));	
		}
		let length= getDistance(x1,y1,x2,y2);  // current tested moved line length
		// test new coord x1, y1, x2, y2. approx range is [-50%, 50%]
		if ( length >= 0.5 * stim[item].fixed.length && length <= 2 * stim[item].fixed.length )  {   // update moved line position	
			stim[item].moved.begin.x= x1;
			stim[item].moved.begin.y= y1;
			stim[item].moved.end.x = x2;
			stim[item].moved.end.y = y2;
			stim[item].moved.radius = r;
			start_rad= cur_rad;

			mis[item]= false;	// this item is not missing
			
			drawStim();  // repaint stimulus	
		}
		// else do nothing (idle)		
	}		 
}


// for precize adjustments you can use arrow Up key (increasing) or arrow Down (decreasing line size).
function onKeyDown(event)
{    
  if (is_task) 
	{
		event = event || window.event;   
		let key = event.keyCode, xc, yc;  
		/*	arrow Left = 37
				arrow Right = 39
				arrow Up = 38
				arrow Down = 40		*/			
		if ( key != 38 && key != 40 )  {
			return;
		}
		// if future todo. this code section is repeating approx 3. times in vhi.js. create new function for bellow section. see mouseMove()
		let fixed_v = fixed_line == '2' ? false: true;  // window.fixed_line: Vertical == '1' or Horizontal == '2'
		let r = key == 38 ? stim[item].moved.radius + 1 : stim[item].moved.radius - 1; 
		if ( fixed_v )  { 
			xc= divLine(stim[item].moved.begin.x, stim[item].moved.end.x, -1);			
			yc= divLine(stim[item].moved.begin.y, stim[item].moved.end.y, -1);		
		}
		else  {
			xc = stim[item].moved.begin.x;  // x-intersection of vertical (moved) and horizontal (fixed) line
			yc = stim[item].moved.begin.y;  // y-intersection of vertical (moved) and horizontal (fixed) line
		}  
		let angle= stim[item].moved.angle, x1, y1, x2, y2;
		if ( fixed_v )  { 
			x1 = Math.round (xc - r * Math.cos(angle));
			y1 = Math.round (yc - r * Math.sin(angle));
			x2=  Math.round (xc + r * Math.cos(angle));
			y2=  Math.round (yc + r * Math.sin(angle));
		}
		else  {
		  x1 = xc;  // xc = stim[item].moved.begin.x;   // x-intersection of vertical (moved) and horizontal (fixed) line
			y1 = yc;  // yc = stim[item].moved.begin.y;  // y-intersection of vertical (moved) and horizontal (fixed) line
			x2=  Math.round (xc - r * Math.cos(angle));
			y2=  Math.round (yc - r * Math.sin(angle));	
		}
		let length= getDistance(x1,y1,x2,y2);  // current tested moved line length

		if ( length >= 0.5 * stim[item].fixed.length && length <= 2 * stim[item].fixed.length )  
		{  // update moved line position	
			stim[item].moved.begin.x= x1;
			stim[item].moved.begin.y= y1;
			stim[item].moved.end.x = x2;
			stim[item].moved.end.y = y2;
			if ( key == 38 ) { 
				stim[item].moved.radius++;
      }
			else  {
				stim[item].moved.radius--;
			}	
			
			mis[item]= false;	// this item is not missing
	
			drawStim();  // repaint stimulus	
		}
		// else idle 
  } 
}


// draw (repaint) current stimulus           
function drawStim() 
{    		 
	let c= document.getElementById('can'), ctx= c.getContext('2d'), w= c.width, h= c.height;			
	
	ctx.fillStyle= back_clr;
	ctx.fillRect(0, 0, w, h);    // ctx.clearRect(0, 0, w, h); 

	ctx.strokeStyle= pen_clr; 
	//ctx.lineWidth= pen_width;  see setCanvas()
		
	ctx.beginPath();
	ctx.moveTo(stim[item].fixed.begin.x, stim[item].fixed.begin.y);  // move to start point of fixed line     
	ctx.lineTo(stim[item].fixed.end.x, stim[item].fixed.end.y);  // draw line to end point of fixed line 
  ctx.stroke(); 

	if (demo)  
	{	 // draw the correct line (solution) => before moved line (on "bottom")
		ctx.strokeStyle= cor_clr;   
		ctx.beginPath(); 
		ctx.moveTo(stim[item].correct.begin.x, stim[item].correct.begin.y);  // move to begin point of moved (scroll) line     
		ctx.lineTo(stim[item].correct.end.x, stim[item].correct.end.y);  // draw line to end point of moved (scroll) line 
		ctx.stroke();			
	} 	

	ctx.strokeStyle= is_move ? drag_clr : pen_clr_m; 
  ctx.beginPath(); 
	ctx.moveTo(stim[item].moved.begin.x, stim[item].moved.begin.y);  // move to start point of moved (scroll) line     
	ctx.lineTo(stim[item].moved.end.x, stim[item].moved.end.y);  // draw line to end point of moved (scroll) line 
	ctx.stroke();  
 
	if ( demo &&  !mis[item] )  {	
		calcError(item);
		ctx.fillStyle= pen_clr;    //ctx.fillStyle= '#000';
		let x= stim[item].moved.x0;  // divLine(stim[item].moved.begin.x, stim[item].moved.end.x, -1);		// x-half of moved line
		let y= stim[item].moved.y0;  // divLine(stim[item].moved.begin.y, stim[item].moved.end.y, -1);		 // y-half of moved line		
		ctx.fillText('Absolute error= ' + dv[item][0].toFixed(3) + '. Relative error= ' + dv[item][1].toFixed(3), x, y);	  
	}		
	
	updateFooter();	
	
}


// calculate dependent variable: absolute and relative error.
// i => item number	
function calcError(i)  
{
	let f_length= getDistance(stim[i].fixed.begin.x, stim[i].fixed.begin.y, stim[i].fixed.end.x, stim[i].fixed.end.y);  // fixed line length (px)
	let m_length= getDistance(stim[i].moved.begin.x, stim[i].moved.begin.y, stim[i].moved.end.x, stim[i].moved.end.y);  // current moved line length (px)
  dv[i][0]= Math.abs(f_length - m_length);  // absolute errror in pixels. Note: always is >= 0 !!
	if ( m_length < f_length )  {  // approx range = [-50..50]
		dv[i][1]= - ( 100 -  ( 100 * ( m_length  / f_length ) ) );  // error in "percents"
  }
	else if ( m_length > f_length )  {
		dv[i][1]=  100 - ( 100 * ( f_length  / m_length ) );  // error in "percents"
	}
	else  {
		dv[i][1]= 0;
	}		
}

	
// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   

	if (document.getElementById('iter').validity.valid)  {  
	  iter = Math.round(parseInt(document.getElementById('iter').value, 10));			
  } 	
	
	if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr').value))  {		
		pen_clr= document.getElementById('pen_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr_m').value))  {		
		pen_clr_m= document.getElementById('pen_clr_m').value;
  }
	
  if (validColor(document.getElementById('drag_clr').value))  {		
		drag_clr= document.getElementById('drag_clr').value;
  }		
	
	if (validColor(document.getElementById('cor_clr').value))  {		
		cor_clr= document.getElementById('cor_clr').value;
  }		
	
	if (document.getElementById('fixed_line').selectedIndex > -1 )  {
		fixed_line= document.getElementById('fixed_line').value;
	}
	
	if (document.getElementById('pen_width').validity.valid)  {  
	  pen_width = Math.round(parseInt(document.getElementById('pen_width').value, 10));			
  }
  
	is_back= document.getElementById('is_back').checked ? true : false;      
  	
	prg_type= document.getElementById('prg_type').checked ? true : false;   
		
  demo= document.getElementById('demo').checked ? true : false;	
	
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {
    item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{    		
	item++;					
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


/* divide line 
  first => first point (P1) 
  second => second point (P2)
  n => (distance(P1-Q)/ distance(Q-P2)) = lambda = -m/n = negative ratio length: P1-Q/Q-P2
  use n= -1 for half divide line (50%) 
	returns: coordinate of point (Q) between first and second point   */   
function divLine(first,second,n)  {     
  return (first-n*second)/(1-n);   // Math.round((first-n*second)/(1-n));
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type)  {    // update Progress bar => show current item count and remaining item count on Progress Bar                                                                                                                                                        
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent= 'Item= ' + (item + 1) + ' Remaining items= ' + (item_cnt - item-1);  
	}
}


// end of task (research)
function theEnd()  
{     
	is_task= false;

	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
  document.getElementById('can').style.display= 'none';   
	document.getElementById('result_wrapper').style.display= 'block';
	
	document.getElementById('warning').style.display= resized == 0 ? 'none' : 'block';
	
	const D = 3;  // number of decimals
	/*
	  blank = number of missing items in task = not moved items
		rows = array of table (table1,table2,..) rows
		length_px= array of length of fixed line for each item in task
		err_a = array of ALL absolute error in task (unit is px)
		err_r = array of ALL relative error in task. range: [-50%..50%]
		var_g => object for group by user defined % length (absolute and relative errors group by user defined line length value in %)    
	*/
	let blank= 0, rows = [], length_px= []; err_a = [], err_r = [], var_g= {};  // non-missing error (px), non-missing error (%),			

	let header= [ 'Item', 'Length-%',  'Length-px', 'Absolute error', 'Relative error', 'Moved' ];
	
	for (let i = 0; i < item_cnt; i++)  
	{					
		length_px[i]= getDistance(stim[i].fixed.begin.x, stim[i].fixed.begin.y, stim[i].fixed.end.x, stim[i].fixed.end.y);  // see calcError() => fixed line length (px)
		let key = line_size[i];  // group by this key. key is user defined line length value in %
		if ( typeof var_g[key]  === 'undefined' )  {  // for decriptive statistics set group by line length variable value (table2 + table3). 
			var_g[key] = {   
				'absolute'  : [],  // absolute error 
				'relative' : []    // relative error 
			}; 
		}
		
		if ( mis[i] == true )  {
			blank++;	
			rows[i]= [ parseInt(i+1,10), line_size[i], length_px[i].toFixed(D), '', '', 'No' ];  // Item, Length-%, Length-px, Absolute error, Relative error, Moved
		}
		else  {
      calcError(i); 
			err_a.push ( dv[i][0] );  // all absolute errors 
		  err_r.push ( dv[i][1] );  // all relative errors 
			rows[i]= [ parseInt(i+1,10), line_size[i], length_px[i].toFixed(D), dv[i][0].toFixed(D), dv[i][1].toFixed(D), 'Yes' ]; // Item, Length-%, Length-px, Absolute error, Relative error, Moved
      var_g[key].absolute.push( dv[i][0] );    // absolute error in pixels 
			var_g[key].relative.push( dv[i][1] );   // relative error in percents [-50%..+50%]		
		}
				
	}  // for (let i = 0; i < item_cnt; i++)
		
  createTableArray(document.getElementById('table1'), rows, header); 
  
	// table2 => Absolute error (px). Descriptive statistics grouped by variable Length and total
	//document.getElementById('table2_title').textContent= 'Absolute error (px). Descriptive statistics grouped by variable Length and total:'; 
	let arr, n, m, sd;
	rows = [];  // reset table rows (for table2)
	for (let key in var_g)  {
		arr= var_g[key].absolute;  // grouped by variable for this key
		n= arr.length;
		m =  n > 0  ?  Stat.m(arr).toFixed(D) : '';      // Errors (px). mean only for non-missing items in task
		sd = n > 1  ?  Stat.sd(arr, m).toFixed(D) : '';  // Errors (px). standard deviation only for non-missing items in task  	 	
		rows.push( [key, m, sd, n] );  
	}
	// totals (px)
	n= err_a.length;  // err_a array => all absolute errors 
	m =   n > 0  ?  Stat.m(err_a).toFixed(D) : '';      // Total Errors (px). mean only for non-missing items in task
	sd =  n > 1  ?  Stat.sd(err_a, m).toFixed(D) : '';  // Total Errors (px). standard deviation only for non-missing items in task  	 	
  rows.push( ['Total (px)', m, sd, n] );  // last row
	let data = [ m, sd ];  // index 0 => mean for all absolute errors, index 1 => stand. deviation for all absolute errors
	header= ['Length', 'Average', 'Standard Deviation', 'N']; // header		
	createTableArray(document.getElementById('table2'), rows, header); 
	
	// table3 => Relative error (%). Descriptive statistics grouped by variable Length and Total:
	//document.getElementById('table3_title').textContent= 'Relative error. Descriptive statistics grouped by variable Length and total:';
	rows = [];  // reset table rows (for table3)
	for (let key in var_g)  {
		arr= var_g[key].relative;  // grouped by variable for this key
		n= arr.length;
		m =  n > 0  ? Stat.m(arr).toFixed(D) : '';      //  Errors (%). mean only for non-missing items in task
		sd = n > 1  ? Stat.sd(arr, m).toFixed(D) : '';  //  Errors (%). standard deviation only for non-missing items in task  	 	
		rows.push( [key, m, sd, n] );  
	}
  // totals	(%)
  n= err_r.length;  // err_r array => all relative errors 
	m =  n > 0  ?  Stat.m(err_r).toFixed(D) : '';     //  Total Errors (%). mean only for non-missing items in task
	sd = n > 1  ?  Stat.sd(err_r, m).toFixed(D) : '';  // Total Errors (%). standard deviation only for non-missing items in task  	 	
  rows.push( ['Total', m, sd, n] );  // last row	
	data.push(m);   // index 2 => mean for all relative errors
	data.push(sd);  // index 3 => stand. all relative errors
	createTableArray(document.getElementById('table3'), rows, header); 
 
	// calculate inferential statistics (pearson r, sperman ro, t-arguments, t-distribution probabilities)	
	n= err_a.length;  // n= err_r.length; 
	header=  ['Variables', 'Spearman-ro', 't-value ro', 'df', 'p-ro (two-sided)',  'Pearson r', 't-value r', 'df', 'p-r (two-sided)' ]; // header 
	rows=[];    // reset table rows (for table4)
	for (let j= 0; j<=1; j++)
	{
		let y=[], vars;
		if ( j == 0 )  {
			y= err_a;   // absolute error in pixels	
			vars= 'Length(px)-Absolute error'; 
		}
		else  {
			y= err_r;    // relative error in range [-50..50]
			vars = 'Length(px)-Relative error'; 
		}	
		let row= [ vars, '', '', '', '', '', '', '', '' ];  // range 0..8 
		if  ( n >= 3 )   // minimum 3 non-missing items
		{			
			let ro= Stat.r_spearman(length_px, y);     // spearman's rank correlation
			if (!isNaN(ro) && isFinite(ro) )  // if no variance in both arrays (sd= 0) => Infinity or Nan
			{ 
				row[1]= ro.toFixed(D);  
				let t_ro= ro * Math.sqrt ( ( n-2 ) / (1 - ro * ro) );  // t-argument for spearman correlation 		
        if (!isNaN(t_ro) && isFinite(t_ro))  
				{
					let err=[0];  // reset error flag indicator (must be 0)
					let p_ro= Stat.prncst ( Math.abs(t_ro), n-2, 0, err[0]); // t-argument, df, 0, error
					p_ro=  2 * (1 - p_ro);  // Propability p (two-sided) => tested ok for agument n= 3..1000 with: https://www.psychometrica.de/correlation.html 
					row[2]= t_ro.toFixed(D);
					row[3]= n-2;
					if (err[0] == 0) { 
						row[4]= p_ro.toFixed(D); // propability Spearman ro (two-sided)
					}
				}
			}
			
			let r= Stat.r(length_px, y);   // Pearson r correlation
			if (!isNaN(r) && isFinite(r))    // if no variance in both arrays (sd= 0) => Infinity or Nan
			{
				row[5]= r.toFixed(D);  
				let t_r=  r * Math.sqrt ( ( n-2 ) / (1 - r * r) );  // t-argument for pearson correlation
				if (!isNaN(t_r) && isFinite(t_r)) 
				{					
					let err=[0];  // reset error flag indicator (must be 0)
					let p_r= Stat.prncst ( Math.abs(t_r), n-2, 0, err[0]);  // t-argument, df, 0, error
					p_r= 2 * (1 - p_r);  // Propability p (two-sided) => tested ok for agument n= 3..1000 with: https://www.psychometrica.de/correlation.html 
					row[6]= t_r.toFixed(D);
					row[7]= n-2;
					if (err[0] == 0) { 
						row[8]= p_r.toFixed(D);  // p-r (two-sided)
					}					 
				}
			}			
		}  //  if ( n >= 3 ) 
		
		rows.push( row );  // 9 x columns
		
		data.push(row[5]);  // pearson r  (index 4 or 7)
		data.push(row[6]);  // t-value r  (index 5 or 8)
		data.push(row[8]);  // p-r (two-sided) => index (6 or 9)
	}		
	
	if ( n >= 3 ) {
	  data.push(n - 2);   // df is equal for px and %. => or: err_a.length - 2 or: n= err_r.length - 2;
	}
  else  {
		data.push('');  // df => minimum if 3 non-missing items
	}
	
	createTableArray(document.getElementById('table4'), rows, header); 
  
	//n= err_a.length;  // non-missing item count (error px)
	document.getElementById('inf_warning').style.display= n >= 3 ? 'block' : 'none';	
	
  // update 'score' label => above "Repeat" button
	switch (n)  {
		case 0: 
			document.getElementById('score').textContent= 'Descriptive statistics cannot be calculated. Missing items= ' + blank;
			break;
		case 1: 
			document.getElementById('score').textContent= 'Total average Absolute error (px)= ' + data[0] + '. Standard deviation cannot be calculated. Missing items= ' + blank;
			break;	
		default:
		document.getElementById('score').textContent= 'Total average Absolute error (px)= ' + data[0] + '. Standard Deviation= ' + data[1] + '. Missing items= ' + blank; 
	}
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	let var_values= '';    // key = line_size[i];  // group by this key. key is user defined line length value in %
  for (let key in var_g)  {
    var_values += key + ';';   
  }
	
	try 
  {
    if (store) {
			setStorage(var_values, blank, data);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  

}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 var_values => list of variable values (user defined values for Line length values for fixed line - %). separator in list is ";"  
	 blank => number of missing (not-moved) items in task 
	 data => descriptive and inferential statistics (index range: 0..10). for details see theEnd()  */ 
function setStorage(var_values, blank, data)  
{    
	let oldItems = JSON.parse(localStorage.getItem('vhi-v1.0')) || []; 
  let id= oldItems.length + 1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
			'iteration' : iter,
			'length-values' : var_values,		
			'fixed': fixed_line, 			
			'pen-w': pen_width,	
			'is-back' : is_back ? 1 : 0,
			'demo' : demo ? 1 : 0, 						
			'items' : item_cnt,
			'missing'  : blank, 
			'resized': resized,
			
			'm-px'  : data[0], 	
			'sd-px' : data[1],
      'm-%'   : data[2], 	
			'sd-%'  : data[3],
			'r-px'  : data[4],
      't-px'  : data[5],
      'p-px'  : data[6],			
			'r-%'   : data[7],
      't-%'   : data[8],
      'p-%'   : data[9],			
			'df'    : data[10]
  };

  oldItems.push(newItem);

  localStorage.setItem('vhi-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
  
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
          col.push(key);
        }
    }
  } 
  let table = document.createElement('table');
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement('th');      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= '';
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('vhi-v1.0')  ?  JSON.parse(localStorage.getItem('vhi-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('vhi-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('vhi-v1.0');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('vhi-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 	
	let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  
			break;
		case 2:
			tbl = document.getElementById('table3').getElementsByTagName('table')[0];  
			break;			
		default:
			tbl = document.getElementById('table4').getElementsByTagName('table')[0];  // note: id='table4' is html div element
			break;	
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}


// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Classes
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Classes/static
// http://www.rosettacode.org/mw/index.php?title=Special%3ASearch&search=integration&go=Go&fulltext=1
class Stat 
{ 
		
	// return arithmetic mean (mean, average) of input array (arr)
  // no validation => input array must be in valid number format	
	static m(arr)  
	{	
		let len= arr.length, sum= 0;
		for (let i=0; i<len; i++) {
			sum+= arr[i];
		}
		return (sum/arr.length);
	}	


	// return Sample Standard Deviation (with Bessel's correction) => N-1
	// no validation => input array must be in valid number format
	// arr = array argument;  mean = average for input arr 
	static sd(arr, mean) 
	{
		let len= arr.length, sum= 0, res;
		for (let i=0; i<len; i++) {
			sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
		}
		res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
		return res;
	}
	
	
	// https://www.psychometrica.de/correlation.html
	// return Pearson correlation coefficient [ Pearson's r, Pearson product-moment correlation coefficient, bivariate correlation ]
	//  x, y = arrays 
	static r(x,y)
	{
		if (x.length != y.length) {
			return;  // undefined
		}
		let cov= 0, n, m1, m2, sd1, sd2;
		n= x.length-1;     
		m1= this.m(x);          //MeanAndStdDev(a, M1, SD1);
		sd1= this.sd(x,m1);
		m2= this.m(y);
		sd2= this.sd(y,m2);    //MeanAndStdDev(b, M2, SD2);		
		for (let i=0; i <= n; i++) {
			cov= ( x[i] - m1 ) * ( y[i] - m2 ) + cov;  // covariance
    }
		return ( cov / ( n * sd1 * sd2 ) );   // denominator = n-1 (Bessel's correction)		
	}
	
	
	// return covariance
  // x, y = arrays   
	static cov(x,y)
	{
		if (x.length != y.length) {
			return;  // undefined
		}
		let cov= 0, n, m1, m2, sd1, sd2;
		n= x.length-1;     
		m1= this.m(x);          
		sd1= this.sd(x,m1);
		m2= this.m(y);
		sd2= this.sd(y,m2);    		
		for (let i=0; i <= n; i++) {
			cov= ( x[i] - m1 ) * ( y[i] - m2 ) + cov  ;  // covarijanca
    }
		return cov;   // denominator = n-1 (Bessel's correction)		
	}


	// https://www.statisticshowto.com/probability-and-statistics/correlation-coefficient-formula/spearman-rank-correlation-definition-calculate/   
	// https://mathcracker.com/spearman-correlation-calculator
	// compute Spearman's rank correlation coefficient
	// x, y = arrays
	static r_spearman(x, y)
	{ 
		if (x.length != y.length) {
			return;  // undefined
		}
		const N = x.length;  // n is the number of observations
		let xr = x.slice();  // must be
		xr= this.rankify(xr, xr.length);   // get (tied) ranks for 1. array values
		
		let yr = y.slice();  // must be
		yr= this.rankify(yr, yr.length);	  // get (tied) ranks for 2. array values
		
		let mx= this.m(xr);  // mean rank for 1. array
		let my= this.m(yr);  // mean rank for 2. array
		
		let sxy= 0, sx= 0, sy= 0;
		for (let i= 0; i < N; i++)  
		{
			sxy+= ( xr[i] - mx ) * ( yr[i] - my); 
			sx+= Math.pow( ( xr[i] - mx ), 2);
			sy+= Math.pow( ( yr[i] - my ), 2);
		}
		sxy= sxy/N;
		sx= sx/N;
    sy= sy/N;
    let res = sxy / Math.sqrt(sx * sy);
		return res;	
	}
	
	
	// https://www.geeksforgeeks.org/rank-elements-array/
	// function to find tied ranks
	// A => input array;  n => input array length
	static rankify(A, n) 
	{	
		let R = [...Array(n)];  // Rank Vector
		// Sweep through all elements in A for each
		// element count the number of less than and
		// equal elements separately in r and s.
		for (let i = 0; i < n; i++) 
		{
			let r = 1, s = 1;
			for (let j = 0; j < n; j++) {
				if (j != i && A[j] < A[i]) r += 1;
				if (j != i && A[j] == A[i]) s += 1;
			}
			// Use formula to obtain rank
			R[i] = parseFloat(r + parseFloat(s - 1) / parseFloat(2));
		}
		//for (let i = 0; i < n; i++)  document.write(parseFloat(R[i]).toFixed(1) + "  ");
		return R;
	}
 
   
	// https://people.sc.fsu.edu/~jburkardt/cpp_src/asa005/asa005.cpp
	// https://people.sc.fsu.edu/~jburkardt/cpp_src/asa005/asa005.html
	// https://people.math.sc.edu/Burkardt/cpp_src/toms291/toms291.cpp
	// https://people.math.sc.edu/Burkardt/c_src/toms291/toms291.html


	//****************************************************************************80

	static alnorm ( x, upper )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    ALNORM computes the cumulative density of the standard normal distribution.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    17 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by David Hill.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    David Hill,
	//    Algorithm AS 66:
	//    The Normal Integral,
	//    Applied Statistics,
	//    Volume 22, Number 3, 1973, pages 424-427.
	//
	//  Parameters:
	//
	//    Input, double X, is one endpoint of the semi-infinite interval
	//    over which the integration takes place.
	//
	//    Input, bool UPPER, determines whether the upper or lower
	//    interval is to be integrated:
	//    .TRUE.  => integrate from X to + Infinity;
	//    .FALSE. => integrate from - Infinity to X.
	//
	//    Output, double ALNORM, the integral of the standard normal
	//    distribution over the desired interval.
	//
	{
		let a1 = 5.75885480458;
		let a2 = 2.62433121679;
		let a3 = 5.92885724438;
		let b1 = -29.8213557807;
		let b2 = 48.6959930692;
		let c1 = -0.000000038052;
		let c2 = 0.000398064794;
		let c3 = -0.151679116635;
		let c4 = 4.8385912808;
		let c5 = 0.742380924027;
		let c6 = 3.99019417011;
		let con = 1.28;
		let d1 = 1.00000615302;
		let d2 = 1.98615381364;
		let d3 = 5.29330324926;
		let d4 = -15.1508972451;
		let d5 = 30.789933034;
		let ltone = 7.0;
		let p = 0.398942280444;
		let q = 0.39990348504;
		let r = 0.398942280385;
		let utzero = 18.66;
		let value;
		let y;
		let z = x;

		let up = upper;   // bool
		z = x;

		if ( z < 0.0 )
		{
			up = !up;
			z = - z;
		}

		if ( ltone < z && ( ( !up ) || utzero < z ) )
		{
			if ( up ) {
				value = 0;
			}
			else  {
				value = 1;
			}
			return value;
		}

		y = 0.5 * z * z;

		if ( z <= con )
		{
			value = 0.5 - z * ( p - q * y 
				/ ( y + a1 + b1 
				/ ( y + a2 + b2 
				/ ( y + a3 ))));
		}
		else
		{ 
			value = r * Math.exp( - y )   
				/ ( z + c1 + d1 
				/ ( z + c2 + d2 
				/ ( z + c3 + d3 
				/ ( z + c4 + d4 
				/ ( z + c5 + d5 
				/ ( z + c6 ))))));
		}

		if ( !up )  {
			value = 1 - value;
		}

		return value;
	}


	//****************************************************************************80

	static prncst ( st, idf, d, ifault )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    PRNCST computes the lower tail of noncentral T distribution.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    26 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by BE Cooper.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    BE Cooper,
	//    Algorithm AS 5:
	//    The Integral of the Non-Central T-Distribution,
	//    Applied Statistics,
	//    Volume 17, Number 2, 1968, page 193.
	//
	//  Parameters:
	//
	//    Input, double ST, the argument.
	//
	//    Input, int IDF, the number of degrees of freedom.
	//
	//    Input, double D, the noncentrality parameter.
	//
	//    Output, int *IFAULT, error flag.
	//    0, no error occurred.
	//    nonzero, an error occurred.
	//
	//    Output, double PRNCST, the value of the lower tail of
	//    the noncentral T distribution.
	//
	//  Local Parameters:
	//
	//    Local, double G1, 1.0 / sqrt(2.0 * pi)
	//
	//    Local, double G2, 1.0 / (2.0 * pi)
	//
	//    Local, double G3, sqrt(2.0 * pi)
	//
	{
		let a, ak, b,  da, drb,  fk, fkm1, fmkm1, fmkm2, ioe, k, rb, sum, value;   
		let emin = 12.5;
		let g1 = 0.3989422804;
		let g2 = 0.1591549431;
		let g3 = 2.5066282746;
		let f = idf;   //( double ) ( idf );
	//
	//  For very large IDF, use the normal approximation.
	//
		if ( 100 < idf )
		{
			ifault = 1;

			a = Math.sqrt ( 0.5 * f ) * Math.exp ( this.lgamma ( 0.5 * ( f - 1 ) ) 
				- this.lgamma ( 0.5 * f ) ) * d;

			value = this.alnorm ( ( st - a ) / Math.sqrt ( f * ( 1 + d * d ) 
				/ ( f - 2 ) - a * a ), false );
			return value;
		}

		ifault = 0;
		ioe = ( idf % 2 );
		a = st / Math.sqrt ( f );
		b = f / ( f + st * st );
		rb = Math.sqrt ( b );
		da = d * a;
		drb = d * rb;

		if ( idf == 1 )
		{
			value = this.alnorm ( drb, true ) + 2 * this.tfn ( drb, a );
			return value;
		}

		sum = 0;

		if ( Math.abs ( drb ) < emin )
		{
			fmkm2 = a * rb * Math.exp ( - 0.5 * drb * drb ) 
			* this.alnorm ( a * drb, false ) * g1;
		}
		else
		{
			fmkm2 = 0;
		}

		fmkm1 = b * da * fmkm2;

		if ( Math.abs ( d ) < emin )
		{
			fmkm1 = fmkm1 + b * a * g2 * Math.exp ( - 0.5 * d * d );
		}

		if ( ioe == 0 )
		{
			sum = fmkm2;
		}
		else
		{
			sum = fmkm1;
		}

		ak = 1;
		fk = 2;

		for ( k = 2; k <= idf - 2; k = k + 2 )
		{
			fkm1 = fk - 1;
			fmkm2 = b * ( da * ak * fmkm1 + fmkm2 ) * fkm1 / fk;
			ak = 1 / ( ak * fkm1 );
			fmkm1 = b * ( da * ak * fmkm2 + fmkm1 ) * fk / ( fk + 1 );

			if ( ioe == 0 )
			{
				sum += fmkm2;
			}
			else
			{
				sum += fmkm1;
			}
			ak = 1 / ( ak * fk );
			fk += 2;
		}

		if ( ioe == 0 )
		{
			value = this.alnorm ( d, true ) + sum * g3;
		}
		else
		{
			value = this.alnorm ( drb, true ) + 2 * ( sum + this.tfn ( drb, a ) );
		}

		return value;
	}
	//****************************************************************************80


	// https://people.math.sc.edu/Burkardt/c_src/toms291/toms291.html
	// https://people.math.sc.edu/Burkardt/cpp_src/toms291/toms291.cpp
	//****************************************************************************80

	static lgamma ( x )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    ALOGAM computes the logarithm of the Gamma function.
	//
	//  Modified:
	//
	//    22 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by Malcolm Pike, David Hill.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    Malcolm Pike, David Hill,
	//    Algorithm 291:
	//    Logarithm of Gamma Function,
	//    Communications of the ACM,
	//    Volume 9, Number 9, September 1966, page 684.
	//
	//  Parameters:
	//
	//    Input, double X, the argument of the Gamma function.
	//    X should be greater than 0.
	//
	//    Output, double lgamma, the logarithm of the Gamma
	//    function of X.
	//
	{
		let f, value, y, z;

		if ( x <= 0 || isNaN(x) )  {
			return NaN;
		}

		y = x;

		if ( x < 7 )
		{
			f = 1;
			z = y;

			while ( z < 7 )
			{
				f = f * z;
				z++; 
			}
			y = z;
			f = - Math.log ( f );
		}
		else
		{
			f = 0;
		}

		z = 1 / y / y;

		value = f + ( y - 0.5 ) * Math.log ( y ) - y 
			+ 0.918938533204673 + 
			((( 
			- 0.000595238095238   * z 
			+ 0.000793650793651 ) * z 
			- 0.002777777777778 ) * z 
			+ 0.083333333333333 ) / y;

		return value;
	}

	//****************************************************************************80

	static tfn ( x, fx )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    TFN calculates the T-function of Owen.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    16 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by JC Young, Christoph Minder.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    MA Porter, DJ Winstanley,
	//    Remark AS R30:
	//    A Remark on Algorithm AS76:
	//    An Integral Useful in Calculating Noncentral T and Bivariate
	//    Normal Probabilities,
	//    Applied Statistics,
	//    Volume 28, Number 1, 1979, page 113.
	//
	//    JC Young, Christoph Minder,
	//    Algorithm AS 76: 
	//    An Algorithm Useful in Calculating Non-Central T and 
	//    Bivariate Normal Distributions,
	//    Applied Statistics,
	//    Volume 23, Number 3, 1974, pages 455-457.
	//
	//  Parameters:
	//
	//    Input, double X, FX, the parameters of the function.
	//
	//    Output, double TFN, the value of the T-function.
	//
	{
		const NG = 5;

		let fxs, i, r1, r2, rt, value, x1, x2, xs;
		let r = [
			0.1477621, 
			0.1346334, 
			0.1095432, 
			0.0747257, 
			0.0333357 ];
		let tp = 0.159155;
		let tv1 = 1.0E-35;
		let tv2 = 15.0;
		let tv3 = 15.0;
		let tv4 = 1.0E-05;
		let u = [
			0.0744372, 
			0.2166977, 
			0.3397048,
			0.4325317, 
			0.4869533 ];


	//
	//  Test for X near zero.
	//
		if ( Math.abs ( x ) < tv1 )
		{
			value = tp * Math.atan ( fx );
			return value;
		}
	//
	//  Test for large values of abs(X).
	//
		if ( tv2 < Math.abs ( x ) )
		{
			value = 0.0;
			return value;
		}
	//
	//  Test for FX near zero.
	//
		if ( Math.abs ( fx ) < tv1 )
		{
			value = 0;
			return value;
		}
	//
	//  Test whether abs ( FX ) is so large that it must be truncated.
	//
		xs = - 0.5 * x * x;
		x2 = fx;
		fxs = fx * fx;
	//
	//  Computation of truncation point by Newton iteration.
	//
		if ( tv3 <= Math.log ( 1 + fxs ) - xs * fxs )
		{
			x1 = 0.5 * fx;
			fxs = 0.25 * fxs;

			for ( ; ; )
			{
				rt = fxs + 1;

				x2 = x1 + ( xs * fxs + tv3 - Math.log ( rt ) ) 
				/ ( 2 * x1 * ( 1 / rt - xs ) );

				fxs = x2 * x2;

				if ( Math.abs ( x2 - x1 ) < tv4 )  {
					break;
				}
				x1 = x2;
			}
		}
	//
	//  Gaussian quadrature.
	//
		rt = 0.0;
		for ( i = 0; i < NG; i++ )
		{
			r1 = 1 + fxs * Math.pow ( 0.5 + u[i], 2 );
			r2 = 1 + fxs * Math.pow ( 0.5 - u[i], 2 );

			rt = rt + r[i] * ( Math.exp ( xs * r1 ) / r1 + Math.exp ( xs * r2 ) / r2 );
		}

		value = rt * x2 * tp;

		return value;

	}
	//****************************************************************************80
	
} 
